"""
Module: '_api.lightmatrix' on micropython-v1.14-893-lego learning system hub
"""
# MCU: {'machine': 'LEGO Technic Large Hub with STM32F413xx', 'sysname': 'LEGO Technic Large Hub', 'platform': 'LEGO Learning System Hub', 'nodename': 'LEGO Learning System Hub', 'ver': 'v1.14-893', 'release': '1.14.0', 'name': 'micropython', 'family': 'micropython', 'port': 'LEGO Learning System Hub', 'version': '1.14.0', 'mpy': 517, 'build': '893'}
# Stubber: 1.7.2
from typing import Any

FACES = [] # type: list

class LightMatrix():
    def __init__(self, *argv, **kwargs) -> None:
        """"""

    def get_orientation(self, *args, **kwargs) -> Any:
        """"""

    def off(self, *args, **kwargs) -> Any:
        """"""

    def play_animation(self, *args, **kwargs) -> Any:
        """"""

    def rotate(self, *args, **kwargs) -> Any:
        """"""

    def set_orientation(self, *args, **kwargs) -> Any:
        """"""

    def set_pixel(self, *args, **kwargs) -> Any:
        """"""

    def show(self, *args, **kwargs) -> Any:
        """"""

    def show_image(self, *args, **kwargs) -> Any:
        """"""

    def start_animation(self, *args, **kwargs) -> Any:
        """"""

    def write(self, *args, **kwargs) -> Any:
        """"""

ORIENTATIONS = [] # type: list
def get_current_orientation(*args, **kwargs) -> Any:
    """"""

def rotate_orientation(*args, **kwargs) -> Any:
    """"""

def set_current_orientation(*args, **kwargs) -> Any:
    """"""

