class Battery:
    """
    A stub class to represent the battery member of the EV3Brick class.
    """

    def __init__(self):
        ...

    def voltage(self) -> int:
        """
        Gets the voltage of the battery.

        Returns:
        Voltage in mV.
        """
        return 0

    def current(self) -> int:
        """
        Gets the current supplied by the battery.

        Returns:
        Battery current in mA.
        """
        return 0
